VERTEX_TYPE_OFFSET0_STEP2: int = 1
VERTEX_TYPE_OFFSET2_STEP5: int = 2
VERTEX_OFFSET: int = 0
VERTEX_STEP: int = 2
VERTEX_TYPE: int = VERTEX_TYPE_OFFSET0_STEP2
REVERSE_TEXTURE_T: bool = True
MAX_INTERPOLATION: int = 5
PIVOT_TABLE_SIZE: int = 65
GOSA: float = 0.0001

LIVE2D_FORMAT_VERSION_V2_8_TEX_OPTION: int = 8
LIVE2D_FORMAT_VERSION_V2_10_SDK2: int = 10
LIVE2D_FORMAT_VERSION_V2_11_SDK2_1: int = 11
LIVE2D_FORMAT_VERSION_AVAILABLE: int = LIVE2D_FORMAT_VERSION_V2_11_SDK2_1
OBJECT_REF = 33
