/**
 * Copyright(c) Live2D Inc. All rights reserved.
 *
 * Use of this source code is governed by the Live2D Open Software license
 * that can be found at https://www.live2d.com/eula/live2d-open-software-license-agreement_en.html.
 */

#pragma once

/**
 * Header to include the Live2D Cubism Core in the `Live2D::Cubism::Core` namespace
 *
 * @note Include this header instead of including the `Live2DCubismCore.h` directly.
 */
namespace Live2D { namespace Cubism { namespace Core {
#include "Live2DCubismCore.h"
}}}
